#include		"Intersect.h"

void clIntersect::funcGetIntervalSphere( clSphere a, clVec2 axis, float & fMin, float & fMax )
{
	fMin = ( a.vPos - axis * a.fRadius ) & axis;
	fMax = ( a.vPos + axis * a.fRadius ) & axis;

	if( fMin > fMax )
	{
		float	fTemp = fMin;
		fMin = fMax;
		fMax = fTemp;
	}
}

void clIntersect::funcGetIntervalLine( clLine a, clVec2 axis, float & fMin, float & fMax )
{
	fMin = a.vVec[ 0 ] & axis;
	fMax = a.vVec[ 1 ] & axis;

	if( fMax < fMin )
	{
		float	fTemp = fMax;
		fMax = fMin;
		fMin = fTemp;
	}
}

void clIntersect::funcGetInterval( clVec2 * v, int iNum, clVec2 vAxis, float & fMin, float & fMax )
{
	fMin = v[ 0 ] & vAxis;
	fMax = v[ 0 ] & vAxis;

	for( int i = 1; i < iNum; i ++ )
	{
		float	fTemp = v[ i ] & vAxis;

		if( fMin > fTemp )
		{
			fMin = fTemp;
		}
		else
		{
			if( fMax < fTemp )
			{
				fMax = fTemp;
			}
		}
	}
}

bool clIntersect::funcIntersectAABBxAABB( clAABB & a0, clAABB & a1 )
{
	clVec2	vDel = a1.vPos - a0.vPos;	//    

	//    vDel   X  Y <=     
	return ( ( fabs( vDel.x ) <= ( a0.vSiz.x + a1.vSiz.x ) ) && 
			 ( fabs( vDel.y ) <= ( a0.vSiz.y + a1.vSiz.y ) ) );
}

bool clIntersect::funcIntersectSpherexSphere( clSphere & a0, clSphere & a1, clVec2 & v, float & fDepth, clVec2 & c )
{
	float	fRadx2 = a0.fRadius + a1.fRadius;	//  
	v = a0.vPos - a1.vPos;		//   
	float	fDist = ( v ) & ( v );	//  

	//           
	if( fRadx2 * fRadx2 >= fDist )
	{
		float	f = sqrt( fDist );
		v /= f;		//  

		//   
		c = a1.vPos + v * a1.fRadius;

		fDepth = ( fRadx2 - f );

		return true;
	}
	else
	{
		v.VecZero();

		return false;
	}
}

bool clIntersect::funcIntersectAABBxSphere( clAABB & a0, clSphere & a1, clVec2 & v, float & fDepth, clVec2 & c )
{
	float	faMin, faMax;
	float	fbMin, fbMax;
	float	fD0, fD1;	//      
	float	fTemp;
	clVec2	vDel = a0.vPos - a1.vPos;	//   

	a0.funcReload();

	//      X  Y
	clVec2	vAxis( 1.0f, 0.0f );
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	v = vAxis;
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	vAxis.VecSet( 0.0f, 1.0f );
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//   4       
	// A B
	// C D
	//  B
	vAxis = a1.vPos - a0.vPos + a0.vSiz;
	vAxis.VecNormalize();
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//  C
	vAxis = a1.vPos - a0.vPos - a0.vSiz;
	vAxis.VecNormalize();
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//  A
	vAxis.VecSet( a1.vPos.x - a0.vPos.x - a0.vSiz.x, a1.vPos.y - a0.vPos.y + a0.vSiz.y );
	vAxis.VecNormalize();
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	//  D
	vAxis.VecSet( a1.vPos.x - a0.vPos.x + a0.vSiz.x, a1.vPos.y - a0.vPos.y - a0.vSiz.y );
	vAxis.VecNormalize();
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	funcGetIntervalSphere( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	// 
	if( ( v & vDel ) > 0.0f )
	{
		v *= - 1;
	}

	c = a1.vPos - v * a1.fRadius;

	return true;
}

bool clIntersect::funcIntersectSpherexLine( clSphere & a0, clLine & a1, clVec2 & v, float & fDepth, clVec2 & c )
{
	float	faMin, faMax;	//  
	float	fbMin, fbMax;	//  
	float	fD0, fD1;	//      
	float	fTemp;
	clVec2	vDel = a0.vPos - a1.vPos;	//   

	//   
	a1.funcReload();

	//    A   B
	//      3 
	// 1     A
	clVec2	vAxis = ( a0.vPos - ( a1.vPos - a1.vDir * a1.fHalf ) );
	vAxis.VecNormalize();
	funcGetIntervalSphere( a0, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || faMax < fbMin )
	{
		return false;
	}

	//   
	v = vAxis;
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	// 2     B
	vAxis = ( a0.vPos - ( a1.vPos + a1.vDir * a1.fHalf ) );
	vAxis.VecNormalize();
	funcGetIntervalSphere( a0, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || ( faMax < fbMin ) )
	{
		return false;
	}

	//   
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	// 3           
	vAxis = a1.vDir.VecNormal();
	vAxis.VecNormalize();
	funcGetIntervalSphere( a0, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( faMin > fbMax ) || ( faMax < fbMin ) )
	{
		return false;
	}

	//   
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;
	}

	if( ( v & vDel ) > 0.0f )
	{
		v *= - 1;
	}

	c = a0.vPos + v * a0.fRadius;

	//    ,   
	return true;
}

bool clIntersect::funcIntersectSpherexOBB( clSphere & a0, clOBB & a1, clVec2 & v, float & fDepth, clVec2 & c )
{
	clSphere	sphere;
	clAABB		aabb;

	clVec2		vDel = a0.vPos - a1.vPos;	//    
	//     ,      OBB
	vDel.VecRotate( - a1.fAngle );
	sphere.vPos = a1.vPos + vDel;	//    
	sphere.fRadius = a0.fRadius;	// 
	//   AABB
	aabb.vPos = a1.vPos;
	aabb.vSiz = a1.vSiz;

	//    AABB x Sphere
	if( funcIntersectAABBxSphere( aabb, sphere, v, fDepth, c ) )
	{
		v.VecRotate( a1.fAngle );

		c = a0.vPos - v * a0.fRadius;

		return true;
	}
	else
	{
		return false;
	}
}

bool clIntersect::funcIntersectAABBxLine( clAABB & a0, clLine & a1, clVec2 & v, float & fDepth, clVec2 & c )
{
	float	faMin, faMax;	//  AABB
	float	fbMin, fbMax;	//  
	float	fD0, fD1;	//      
	float	fTemp;
	clVec2	vDel = a0.vPos - a1.vPos;	//   

	a0.funcReload();
	a1.funcReload();

	//     -1,     
	int		iEdge0 = - 1;
	int		iEdge1 = - 1;

	//  AABB      
	// fbMin -          
	clVec2	vAxis = a1.vDir.VecNormal();
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	fbMin = a1.vPos & ( vAxis );
	if( ( fbMin < faMin ) || ( fbMin > faMax ) )
	{
		return ( false );
	}

	v = vAxis;
	fD0 = faMax - fbMin;
	fD1 = fbMin - faMin;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	//    X
	vAxis.VecSet( 1.0f, 0.0f );
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( fbMax < faMin ) || ( fbMin > faMax ) )
	{
		return ( false );
	}

	//   
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;

		iEdge0 = 2;
		iEdge1 = 3;
	}

	//    Y
	vAxis.VecSet( 0.0f, 1.0f );
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	funcGetIntervalLine( a1, vAxis, fbMin, fbMax );
	if( ( fbMax < faMin ) || ( fbMin > faMax ) )
	{
		return ( false );
	}

	//   
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;

		iEdge0 = 1;
		iEdge1 = 2;
	}

	//         
	if( ( vDel & v ) > 0.0f ) 
	{
		v *= - 1;

		//     
		if( iEdge0 != - 1 )	//    
		{
			if( iEdge0 == 1 )
			{
				iEdge0 = 3;
				iEdge1 = 0;
			}
			else
			{
				iEdge0 = 0;
				iEdge1 = 1;
			}
		}
	}

	//   
	if( iEdge0 != - 1 )
	{
		c = this->funcGetFarPoint( a1.vVec, 2, a0.vVec[ iEdge0 ], a0.vVec[ iEdge1 ], v );
	}
	else	//   
	{
		c = this->funcGetFarPoint( a0.vVec, 4, a1.vVec[ 0 ], a1.vVec[ 1 ], v.GetInverse() );
	}

	return ( true );
}

bool clIntersect::funcIntersectOBBxLine( clOBB & a0, clLine & a1, clVec2 & v, float & fDepth, clVec2 & c )
{
	clAABB		aabb;
	clLine		line;

	aabb.vPos = a0.vPos;
	aabb.vSiz = a0.vSiz;

	//       OBB
	clVec2		vDel = a1.vPos - a0.vPos;
	vDel.VecRotate( - a0.fAngle );
	line.vPos = a0.vPos + vDel;
	
	//  
	vDel = a1.vDir;
	vDel.VecRotate( - a0.fAngle );
	line.vDir = vDel;

	//    
	line.fHalf = a1.fHalf;

	if( funcIntersectAABBxLine( aabb, line, v, fDepth, c ) )
	{
		//       
		v.VecRotate( a0.fAngle );

		clVec2	vDel = c - a0.vPos;
		vDel.VecRotate( a0.fAngle );
		c = a0.vPos + vDel;

		return ( true );
	}
	else
	{
		return ( false );
	}

	return true;
}

bool clIntersect::funcIntersectLinexLine( clLine & a0, clLine & a1, clVec2 & v, float & fDepth, clVec2 & c )
{
	float	fMin, fMax;
	float	fD0, fD1;	//      
	float	fTemp;
	clVec2	vDel = a0.vPos - a1.vPos;	//   
	int		iLine = 0;

	//  
	a0.funcReload();
	a1.funcReload();

	//     a1
	clVec2	vAxis = a1.vDir.VecNormal();
	vAxis.VecNormalize();
	funcGetIntervalLine( a0, vAxis, fMin, fMax );
	float	fDot = a1.vPos & vAxis;
	if( ( fDot < fMin ) || ( fDot > fMax ) )
	{
		return false;
	}

	// 
	v = vAxis;
	fD0 = fDot - fMin;
	fD1 = fMax - fDot;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	//     a0
	vAxis = a0.vDir.VecNormal();
	vAxis.VecNormalize();
	funcGetIntervalLine( a1, vAxis, fMin, fMax );
	fDot = a0.vPos & vAxis;
	if( ( fDot < fMin ) || ( fDot > fMax ) )
	{
		return false;
	}

	// 
	fD0 = fDot - fMin;
	fD1 = fMax - fDot;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;

		iLine = 1;
	}

	//  
	if( ( vDel & v ) > 0.0f ) 
        v *= - 1;

	if( !iLine )
	{
		c = this->funcGetFarPoint( a0.vVec, 2, a0.vVec[ 0 ], a0.vVec[ 1 ], v.GetInverse() );
	}
	else
	{
		c = this->funcGetFarPoint( a1.vVec, 4, a1.vVec[ 0 ], a1.vVec[ 1 ], v );
	}

	return true;
}

bool clIntersect::funcIntersectAABBxOBB( clAABB & a0, clOBB & a1, clVec2 & v, float & fDepth, clVec2 & c )
{
	float	faMin, faMax;	//  AABB
	float	fbMin, fbMax;	//  
	float	fD0, fD1;	//      
	float	fTemp;
	clVec2	vDel = a0.vPos - a1.vPos;	//   

	//   
	a0.funcReload();
	a1.funcReload();

	int		iFirst = 1;
	int		iEdge0 = 2;
	int		iEdge1 = 3;

	//     AABB,    1, 0  0, 1      
	clVec2	vAxis( 1.0f, 0.0f );
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	this->funcGetInterval( a1.vVec, 4, vAxis, fbMin, fbMax );

	if( ( faMax < fbMin ) || ( faMin > fbMax ) )
	{
		return false;
	}

	// 
	v = vAxis;
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fDepth = ( fD0 < fD1 ? fD0 : fD1 );

	vAxis.VecSet( 0.0f, 1.0f );
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	this->funcGetInterval( a1.vVec, 4, vAxis, fbMin, fbMax );

	if( ( faMax < fbMin ) || ( faMin > fbMax ) )
	{
		return false;
	}

	// 
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;

		iEdge0 = 1;
		iEdge1 = 2;

		iFirst = 1;
	}

	//    OBB

	float		fCos = cos( a1.fAngle );
	float		fSin = sin( a1.fAngle );

	vAxis.VecSet( fCos, fSin );
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	this->funcGetInterval( a1.vVec, 4, vAxis, fbMin, fbMax );

	if( ( faMax < fbMin ) || ( faMin > fbMax ) )
	{
		return false;
	}

	// 
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;

		iEdge0 = 0;
		iEdge1 = 1;

		iFirst = 0;
	}

	vAxis.VecSet( - fSin, fCos );
	this->funcGetInterval( a0.vVec, 4, vAxis, faMin, faMax );
	this->funcGetInterval( a1.vVec, 4, vAxis, fbMin, fbMax );

	if( ( faMax < fbMin ) || ( faMin > fbMax ) )
	{
		return false;
	}

	// 
	fD0 = faMax - fbMin;
	fD1 = fbMax - faMin;
	fTemp = ( fD0 < fD1 ? fD0 : fD1 );
	if( fDepth > fTemp )
	{
		fDepth = fTemp;
		v = vAxis;

		iEdge0 = 3;
		iEdge1 = 0;

		iFirst = 0;
	}

	// It's done
	if( ( vDel & v ) > 0.0f )
	{
		v *= - 1;

		if( !iFirst )
		{
			if( iEdge0 == 0 )
			{
				iEdge0 = 2;
				iEdge1 = 3;
			}
			else
			{
				iEdge0 = 1;
				iEdge1 = 2;
			}
		}
		else
		{
			if( iEdge0 == 2 )
			{
				iEdge0 = 0;
				iEdge1 = 1;
			}
			else
			{
				iEdge0 = 3;
				iEdge1 = 0;
			}
		}
	}

	if( !iFirst )
	{
		c = ( this->funcGetFarPoint( a0.vVec, 4, a1.vVec[ iEdge0 ], a1.vVec[ iEdge1 ], v.GetInverse() ) );
	}
	else
	{
		c = ( this->funcGetFarPoint( a1.vVec, 4, a0.vVec[ iEdge0 ], a0.vVec[ iEdge1 ], v ) );
	}

	return true;
}

bool clIntersect::funcIntersectOBBxOBB( clOBB & a0, clOBB & a1, clVec2 & v, float & fDepth, clVec2 & c )
{
	clVec2	vDel = a1.vPos - a0.vPos;

	// AABB   OBB
	clAABB	aabb;
	aabb.vPos = a0.vPos;
	aabb.vSiz = a0.vSiz;

	//   OBB    
	clOBB	obb;
	vDel.VecRotate( - a0.fAngle );
	obb.fAngle = a1.fAngle - a0.fAngle;
	obb.vPos = a0.vPos + vDel;
	obb.vSiz = a1.vSiz;

	if( funcIntersectAABBxOBB( aabb, obb, v, fDepth, c ) )
	{
		v.VecRotate( a0.fAngle );

		clVec2	vDel = c - a0.vPos;
		vDel.VecRotate( a0.fAngle );
		c = a0.vPos + vDel;

		return true;
	}
	else
	{
		return false;
	}
}

//          vA, vB
clVec2 clIntersect::funcGetFarPoint( clVec2 * v, int iNum, clVec2 & vA, clVec2 & vB, clVec2 & vNormal )
{
	int			iMinDot = 0;

	for( int i = 1; i < iNum; i ++ )
	{
		if( ( vNormal & v[ iMinDot ] ) > ( vNormal & v[ i ] ) )
		{
			iMinDot = i;
		}
	}

	// 
	clVec2		r = v[ iMinDot ];

	//   ,     
	//   X
	if( vA.x < vB.x )
	{
		if( r.x < vA.x )
		{
			r.x = vA.x;
		}
		if( r.x > vB.x )
		{
			r.x = vB.x;
		}
	}
	else
	{
		if( r.x < vB.x )
		{
			r.x = vB.x;
		}
		if( r.x > vA.x )
		{
			r.x = vA.x;
		}
	}
	//   Y
	if( vA.y < vB.y )
	{
		if( r.y < vA.y )
		{
			r.y = vA.y;
		}
		if( r.y > vB.y )
		{
			r.y = vB.y;
		}
	}
	else
	{
		if( r.y < vB.y )
		{
			r.y = vB.y;
		}
		if( r.y > vA.y )
		{
			r.y = vA.y;
		}
	}

	return ( r );
}

//    
clVec2 clIntersect::funcProjectPointOnLine( clVec2 & vInput, clVec2 & vA, clVec2 & vB )
{
	clVec2		vDel = vB - vA;
	float		fDot = ( vInput - vA ) & vDel;

	fDot /= ( vDel & vDel );

	return ( vA + vDel * fDot );
}
