#include "Vector.h"

#include <math.h>

clVec2::clVec2( void )
{
	x = y = 0.0f;
}

clVec2::clVec2( float fx, float fy )
{
	x = fx;
	y = fy;
}

void clVec2::operator += ( clVec2 vVector )
{
	x += vVector.x;
	y += vVector.y;
}

void clVec2::operator -= ( clVec2 vVector )
{
	x -= vVector.x;
	y -= vVector.y;
}

void clVec2::operator *= ( clVec2 vVector )
{
	x *= vVector.x;
	y *= vVector.y;
}

void clVec2::operator /= ( clVec2 vVector )
{
	x /= vVector.x;
	y /= vVector.y;
}

void clVec2::operator += ( float fValue )
{
	x += fValue;
	y += fValue;
}

void clVec2::operator -= ( float fValue )
{
	x -= fValue;
	y -= fValue;
}

void clVec2::operator *= ( float fValue )
{
	x *= fValue;
	y *= fValue;
}

void clVec2::operator /= ( float fValue )
{
	x /= fValue;
	y /= fValue;
}

clVec2 clVec2::operator + ( clVec2 vVector )
{
	return clVec2( x + vVector.x, y + vVector.y );
}

clVec2 clVec2::operator - ( clVec2 vVector )
{
	return clVec2( x - vVector.x, y - vVector.y );
}

clVec2 clVec2::operator * ( clVec2 vVector )
{
	return clVec2( x * vVector.x, y * vVector.y );
}

clVec2 clVec2::operator / ( clVec2 vVector )
{
	return clVec2( x / vVector.x, y / vVector.y );
}

clVec2 clVec2::operator + ( float fValue )
{
	return clVec2( x + fValue, y + fValue );
}

clVec2 clVec2::operator - ( float fValue )
{
	return clVec2( x - fValue, y - fValue );
}

clVec2 clVec2::operator * ( float fValue )
{
	return clVec2( x * fValue, y * fValue );
}

clVec2 clVec2::operator / ( float fValue )
{
	return clVec2( x / fValue, y / fValue );
}

float clVec2::operator & ( clVec2 vVector )
{
	return ( x * vVector.x + y * vVector.y );
}

float clVec2::operator ^ ( clVec2 vVector )
{
	return ( this->x * vVector.y - this->y * vVector.x );
}

void clVec2::VecSet( float fx, float fy )
{
	x = fx;
	y = fy;
}

clVec2 clVec2::VecNormal( void )
{
	return clVec2( - this->y, this->x );
}

void clVec2::VecZero( void )
{
	x = y = 0.0f;
}

void clVec2::VecNormalize( void )
{
	float	fLength = sqrt( x * x + y * y );

	if( fLength )
	{
		x /= fLength;
		y /= fLength;
	}
}

float clVec2::CosBetween( clVec2 vVector )
{
	float	fDelx = x - vVector.x;
	float	fDely = y - vVector.y;

	float	fLength = sqrt( fDelx * fDelx + fDely * fDely );

	if( fLength )
		return ( ( x * vVector.x + y * vVector.y ) / fLength );
	else
		return ( 0.0f );
}

clVec2 clVec2::PerpVector( void )
{
	return clVec2( - y, x );
}

clVec2 clVec2::GetNormal( void )
{
	float	fLength  = 1 / sqrt( x * x + y * y );

	return clVec2( - y * fLength, x * fLength );
}

clVec2 clVec2::GetInverse( void )
{
	return clVec2( - this->x, - this->y );
}

float clVec2::GetLength( void )
{
	return sqrt( x * x + y * y );
}

float clVec2::GetQLength( void )
{
	return (x * x + y * y);
}

void clVec2::VecReflect( clVec2 vVector )
{
	float		fDot = ( * this ) & vVector;

	x -= vVector.x * 2 * fDot;
	y -= vVector.y * 2 * fDot;
}

void clVec2::VecRotate( float fAngle )
{
	float		t = x;
	float		fSin = sin( fAngle );
	float		fCos = cos( fAngle );

	x = x * fCos - y * fSin;
	y = t * fSin + y * fCos;
}